
#include "connection.hpp"
#include "event.hpp"
#include "ircmsg.hpp"
#include "linebuffer.hpp"
#include "settings.hpp"
#include "write_irc.hpp"

#include <boost/asio.hpp>

#include <algorithm>
#include <chrono>
#include <fstream>
#include <coroutine>
#include <iostream>
#include <limits>
#include <list>
#include <memory>
#include <string>
#include <string_view>
#include <tuple>
#include <utility>
#include <variant>
#include <vector>
#include <unordered_map>
#include <unordered_set>

#include "ping_thread.hpp"
#include "registration_thread.hpp"
#include "self_thread.hpp"
#include "snote_thread.hpp"

using namespace std::chrono_literals;

auto start(boost::asio::io_context & io, Settings const& settings) -> void
{
    auto const connection = std::make_shared<Connection>(io);

    RegistrationThread::start(*connection, settings.password, settings.username, settings.realname, settings.nickname);
    PingThread::start(*connection);
    SelfThread::start(*connection);
    auto const snote_thread = SnoteThread::start(*connection);

/*
    WatchdogThread::start(*connection);
    IrcParseThread::start(*connection);
    CommandThread::start(*connection);
    auto const priv_thread = PrivThread::start(*connection, "privs.toml");
    echo_thread(*connection, priv_thread);

*/
    snote_thread->sig_snote.connect([](auto tag, auto &match) {
        std::cout << "SNOTE " << static_cast<int>(tag) << std::endl;
        for (auto c : match.get_results())
        {
            std::cout << "    " << std::string_view{c.first, c.second} << std::endl;
        }
    });

    boost::asio::co_spawn(
        io,
        connection->connect(io, settings.host, settings.service),
        [&io, &settings](std::exception_ptr e)
        {
            auto timer = std::make_shared<boost::asio::steady_timer>(io);

            timer->expires_after(5s);
            timer->async_wait([&io, &settings, timer](auto) {
                start(io, settings);
            });
        });
}

auto get_settings() -> Settings
{
    if (auto config_stream = std::ifstream {"config.toml"})
    {
        return Settings::from_stream(config_stream);
    }
    else
    {
        std::cerr << "Unable to open config.toml\n";
        std::exit(1);
    }
}

auto main() -> int
{
    auto const settings = get_settings();
    auto io = boost::asio::io_context{};
    start(io, settings);
    io.run();
}
